package gridsim.gssim.policy;

import eduni.simjava.Sim_system;
import gridsim.Gridlet;
import gridsim.gssim.GssimConstants;
import gridsim.gssim.ResourceHistoryItem;
import gridsim.gssim.ResourceUnitsManagerImpl;
import gridsim.gssim.SubmittedTask;
import gssim.schedframe.scheduling.queues.AbstractStatsSupportingQueue;

import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.Map;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.joda.time.DateTime;

import schedframe.resources.ResourceDescription;
import schedframe.resources.ResourceStateDescription;
import schedframe.resources.units.Processor;
import schedframe.resources.units.Processors;
import schedframe.resources.units.ResourceUnit;
import schedframe.scheduling.TaskInterface;
import schedframe.scheduling.TimeResourceAllocation;
import schedframe.scheduling.utils.ResourceParameterName;

/**
 * 
 * @author Marcin Krystek
 *
 */
public class InExecuionList extends AbstractStatsSupportingQueue<SubmittedTask> {

	private static final long serialVersionUID = 3918519811537322812L;
	private static Log log = LogFactory.getLog(InExecuionList.class);
	
	protected AbstractAllocationPolicy allocPolicy;

	
	public InExecuionList(AbstractAllocationPolicy allocationPolicy){
		this.allocPolicy = allocationPolicy;
		this.measureName = GssimConstants.USAGE_MEASURE_NAME;
		this.stats = allocationPolicy.get_stat();
	}
	
	
	protected void updateStats() {
		if(stats == null)
			return;
		
		double time = Sim_system.clock();
		double load = allocPolicy.calculateTotalLoad(size());

		if(log.isInfoEnabled())
			log.info(measureName + ": " + load);
		
		stats.update(measureName, load, lastUpdateTime); //continuous
		lastUpdateTime = time;
	}
	
	public boolean add(SubmittedTask task) {
		
		if (task.getGridletStatus() != Gridlet.PAUSED) {
			//do not allocate resources for a task, that has been paused
			if(log.isDebugEnabled())
				try {
					log.debug("in executioin, cpu request " + task.getCpuCntRequest());
				} catch (NoSuchFieldException e) {
					e.printStackTrace();
				}
				
			double time = allocPolicy.execute(task);
			if(log.isDebugEnabled()){
				log.debug(task.getJobId() + "_" + task.getId() +
						" starts executing on " + new DateTime() + 
						" will finish after " + time);
			}
			if (time < 0.0)
				return false;
		}
		boolean result = super.add(task);
		if (!result) {
			List<ResourceHistoryItem> usedResources = task.getUsedResources();
			allocPolicy.getAllocationManager().freeResources(usedResources.get(usedResources.size() - 1).getResourceUnits());
		}
		task.setGridletStatus(Gridlet.INEXEC);
		allocPolicy.updatePowerUsage();
		
		
		return result;
	}



	public void add(int index, SubmittedTask element) {
		throw new RuntimeException("Not implemented.");
	}



	public boolean addAll(Collection<? extends SubmittedTask> c) {
		throw new RuntimeException("Not implemented.");
	}



	public boolean addAll(int index, Collection<? extends SubmittedTask> c) {
		throw new RuntimeException("Not implemented.");
	}



	public void clear() {
		throw new RuntimeException("Not implemented.");
	}


	public boolean remove(Object task) {
		if (! (task instanceof SubmittedTask))
			return false;
		
		SubmittedTask st = (SubmittedTask) task;

		boolean ret = false;
		if((ret = super.remove(task))) {
			//do not free resources of a task, that has been paused
			List<ResourceHistoryItem> lastUsedList = st.getUsedResources();
			Map<ResourceParameterName, ResourceUnit> lastUsed = lastUsedList.get(lastUsedList.size() - 1).getResourceUnits();
			allocPolicy.getAllocationManager().freeResources(lastUsed);
			allocPolicy.updatePowerUsage();
			allocPolicy.requestedTimeExpiredHandler.cancel(st);
		}
		
		return ret;
	}



	public SubmittedTask remove(int index) {
		SubmittedTask task = super.remove(index);
		
		List<ResourceHistoryItem> lastUsedList = task.getUsedResources();
		Map<ResourceParameterName, ResourceUnit> lastUsed = lastUsedList.get(lastUsedList.size() - 1).getResourceUnits();
		allocPolicy.getAllocationManager().freeResources(lastUsed);
		allocPolicy.updatePowerUsage();
		
		return task;
	}



	public boolean removeAll(Collection<?> c) {
		throw new RuntimeException("Not implemented.");
	}



	public boolean retainAll(Collection<?> c) {
		throw new RuntimeException("Not implemented.");
	}




	public DateTime getArrivalTime(int pos) throws IndexOutOfBoundsException {
		double d = get(pos).getGridlet().getExecStartTime();
		long l = Double.valueOf(d).longValue() * 1000;
		DateTime time = new DateTime(l);
		return time;
	}

	public String getName() {
		return "In execution task list";
	}

	public int getPriority() {
		return 0;
	}

	public void setName(String name) {
	}

	public void setPriority(int priority) {
	}

	public boolean supportReservations() {
		return false;
	}

	
	 public ArrayList<TimeResourceAllocation> nonARTasksResourceUsageList(ResourceUnit resourceUnit, DateTime startTime, DateTime endTime)throws NoSuchFieldException{
		    ArrayList<TimeResourceAllocation> v = new ArrayList<TimeResourceAllocation>();
			for(int i = 0; i < this.size(); i++) {
				TaskInterface<?> task = this.get(i);
				if(task.getStatus() == Gridlet.INEXEC){
					SubmittedTask rgl = (SubmittedTask) task;
					if(rgl.getReservationID() != -1)
						continue;
					DateTime currentTime = new DateTime();
					long expectedRuntime = ((ARAllocationPolicy)allocPolicy).getExpectedRuntime(task);
					if(expectedRuntime <= 0)
						continue;
					DateTime estimatedFinishTime = new DateTime(currentTime.getMillis() + expectedRuntime * 1000);
					
					ResourceStateDescription resDesc = new ResourceStateDescription(this.allocPolicy.provider);
					resDesc.addResourceUnit(new Processors(this.allocPolicy.provider.getProviderId(), ((ResourceUnitsManagerImpl)allocPolicy.getResourceUnitsManager()).getNumPE(), rgl.getUsedResources().get(rgl.getUsedResources().size() - 1).getResourceUnits().get(ResourceParameterName.CPUCOUNT).getUsedAmount()));
					
					TimeResourceAllocation allocation = new TimeResourceAllocation(currentTime,	estimatedFinishTime);
					allocation.setAllocatedResource(resDesc);
					
			        TimeResourceAllocation executionSlot = new TimeResourceAllocation(allocation);
			        // execution slot ends before requested period starts
			        if (executionSlot.getEnd().getMillis() <= startTime.getMillis())
			          continue;

			        // execution slot starts after requested period ends
			        if (executionSlot.getStart().getMillis() >= endTime.getMillis())
			          break;
			        
			        //slot partially before requested period
			        if (executionSlot.getStart().isBefore(startTime)){
			        	executionSlot = new TimeResourceAllocation(executionSlot);
			        	executionSlot.setStart(startTime);
			        }
			        
			        // first execution slot starts after begin of the requested time period, which means that
			        // resource between start time of requested time period and start time of first task
			        // is not used.
			        if (v.size() == 0)
			        	if (executionSlot.getStart().isAfter(startTime)) {
			        		ResourceUnit ru = (ResourceUnit)resourceUnit.clone();
			        		ru.reset();
			        		ResourceStateDescription rd = new ResourceStateDescription(this.allocPolicy.provider);
			        		rd.addResourceUnit(ru);
			        		v.add(new TimeResourceAllocation(rd, startTime, executionSlot.getStart()));

			        	}
			        
			        //slot partially after requested period
			        TimeResourceAllocation oldSlot;
			        if (executionSlot.getEnd().isAfter(endTime)) {
			        	oldSlot = executionSlot;
			        	executionSlot = new TimeResourceAllocation(oldSlot);
			        	executionSlot.setEnd(endTime);
			        }

			        v.add(executionSlot);
					
				}
			}
			  
		    //if there is no resource usage in the requested period
		    if (v.size() == 0) {
		    	ResourceUnit ru = (ResourceUnit) resourceUnit.clone();
		    	ru.reset();
		    	ResourceStateDescription rd = new ResourceStateDescription(this.allocPolicy.provider);
		    	rd.addResourceUnit(ru);
		    	v.add(new TimeResourceAllocation(rd, startTime, endTime));
		    }
		    else { //last slot ends before end of the requested period
		    	TimeResourceAllocation slot = (TimeResourceAllocation)v.get(v.size() - 1); //last slot
		    	if (slot.getEnd().isBefore(endTime)) {
		    		ResourceUnit ru = (ResourceUnit) resourceUnit.clone();
		    		ru.reset();
		    		ResourceStateDescription rd = new ResourceStateDescription(this.allocPolicy.provider);
		    		rd.addResourceUnit(ru);
		    		v.add(new TimeResourceAllocation(rd, slot.getEnd(), endTime));
		    	}
		    }
		    
		    //Find gaps between slots and overlapping slots, and based on them generate additional slots  
		    int i = 0;
		    while (i < v.size()) {
		      TimeResourceAllocation previousSlot = (TimeResourceAllocation)v.get(i);

		      TimeResourceAllocation slot = (TimeResourceAllocation)v.get(i);
		      
		      if (i + 1 < v.size())
		        slot = (TimeResourceAllocation)v.get(i + 1);
		      else 
		      	break;

		      if(slot.getStart().compareTo(previousSlot.getEnd()) == 0)	// one slot immediately after another
		        i++;
		      else if (slot.getStart().isAfter(previousSlot.getEnd())) { // there is a gap between two slots

		    	ResourceUnit resUsage = (ResourceUnit) resourceUnit.clone();
		    	resUsage.reset();

		    	ResourceStateDescription rd = new ResourceStateDescription(this.allocPolicy.provider);
		    	rd.addResourceUnit(resUsage);
		    	TimeResourceAllocation slot1 = new TimeResourceAllocation(rd, previousSlot.getEnd(), slot.getStart());
		    	v.add(++i, slot1);
		    	
		      } else { 									// slots overlap (slot.getStartTime() < lastSlot.getEndTime())
		        DateTime start = previousSlot.getStart();
		        DateTime newStart = slot.getStart();
		        long lnewEnd = Math.min(previousSlot.getEnd().getMillis(),slot.getEnd().getMillis());
		        DateTime newEnd = new DateTime(lnewEnd);
		        
		        long lend = Math.max(previousSlot.getEnd().getMillis(),slot.getEnd().getMillis());
		        DateTime end = new DateTime(lend);
		        
		        ResourceDescription resUsage = (ResourceDescription) previousSlot.getAllocatedResource().clone();
		        TimeResourceAllocation slot1 = new TimeResourceAllocation(resUsage, start, newStart);

		        resUsage = (ResourceDescription) previousSlot.getAllocatedResource().clone();
		        // mark processors which are unavailable in slot.resourceUnit as unavailable also in resUsage.resourceUnit object.
		        // this should be a synonym for: previousSlot.getResourceUnit().getAmount()+slot.getResourceUnit().getAmount()
		        {
		        	ResourceUnit unit = resUsage.getResourceUnit(ResourceParameterName.CPUCOUNT);
		        	if(unit instanceof List){
		        		List<Processor> peResUsage = (List<Processor>) unit;
		        		ResourceUnit resUnit = slot.getAllocatedResource().getResourceUnit(ResourceParameterName.CPUCOUNT);
		        		if(resUnit instanceof List){
		        			List<Processor> disResUnit = (List<Processor>) resUnit;
		        			
				        	for(int j = 0; j < disResUnit.size(); j++){
				        		Processor p = disResUnit.get(j);
				        		if(p.getStatus() == Processor.Status.BUSY){
				        			peResUsage.get(j).setStatus(Processor.Status.BUSY);
				        		} else if(p.getStatus() == Processor.Status.RESERVED){
				        			peResUsage.get(j).setStatus(Processor.Status.RESERVED);
				        		}
				        	}
		        		} else {
		        			log.warn("CHECK AND VERIFY IF THIS CODE IS CORRECT");
		        			int cnt = Float.valueOf(resUnit.getAmount()).intValue();
		        			for(int j = 0; j < peResUsage.size() && cnt > 0; j++){
		        				if(peResUsage.get(j).getStatus() == Processor.Status.FREE){
		        					peResUsage.get(j).setStatus(Processor.Status.BUSY);
		        					cnt--;
		        				}
		        			}
		        		}
		        
		        	} else {
		        		ResourceUnit prevResUnit = previousSlot.getAllocatedResource().getResourceUnit(ResourceParameterName.CPUCOUNT);
		        		ResourceUnit resUnit = slot.getAllocatedResource().getResourceUnit(ResourceParameterName.CPUCOUNT);
		        		ResourceUnit usedUnit = resUsage.getResourceUnit(ResourceParameterName.CPUCOUNT);
		        		usedUnit.setUsedAmount(prevResUnit.getUsedAmount() + resUnit.getUsedAmount());
		        	}
		        }
		        TimeResourceAllocation slot2 = new TimeResourceAllocation(resUsage, newStart, newEnd);
		        
		        if (previousSlot.getEnd().isAfter(slot.getEnd())) {

		        	resUsage = (ResourceDescription) previousSlot.getAllocatedResource().clone();
		        
		        } else {

		        	resUsage = (ResourceDescription) slot.getAllocatedResource().clone();
		        }
		        
		        TimeResourceAllocation slot3 = new TimeResourceAllocation(resUsage, newEnd, end);
		        
		        if (newEnd.compareTo(end) != 0){
		        	int j=i+2;
		        	while ((j < v.size()) && (newEnd.isAfter(((TimeResourceAllocation)v.get(j)).getStart())))
		        		j++;
		            v.add(j, slot3);
		        }
		        v.remove(previousSlot);
		        v.remove(slot);
		        if (start.compareTo(newStart) != 0) {
			          v.add(i++, slot1);
			          v.add(i, slot2);
			    }
		        else {//equal start times of slots
		        	
		        	TimeResourceAllocation tra = null;
		        	float amount = -1;
		        	if (i > 0) {
		        		tra = (TimeResourceAllocation)v.get(i-1);
		           		amount = tra.getAllocatedResource().getResourceUnit(ResourceParameterName.CPUCOUNT).getUsedAmount();	
		        	}
		        		
		        	if (amount == slot2.getAllocatedResource().getResourceUnit(ResourceParameterName.CPUCOUNT).getUsedAmount()) { //two subsequent slots with equal resource usage
		        		tra.setEnd(slot2.getEnd()); //extend previous slot
		        		i--;
		        	}
		        	else
		        		v.add(i, slot2);
		        }
		      }
		    }

		    if (v.size() == 0){
		    	if(log.isErrorEnabled())
					log.error(getClass().getName()+" resourceUsage() - end, v.size = 0");
		      return null;
		    }
			return v;
		}
	 
}
